# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Number to string and vice versa."""

import re

import numpy as np
import numpy.typing as npt

###############################################################################


class NumStr:
    """Analyse the format of a string representing a number."""

    def __init__(self, input_string, dec_sep="."):
        """
        Analyse the format of a string representing a number.

        Parameters
        ----------
        input_string : str
            String, representing a number.
        dec_sep : str, optional
            Decimal separator. The default is ".".

        Returns
        -------
        None.

        """
        self.input_string = input_string
        self.dec_sep = dec_sep
        self.regexes = {
            "dec": f"[+-]?[0-9]+[{self.dec_sep}][0-9]*|[+-]?[0-9]*[{self.dec_sep}][0-9]+",
            "no_dec": "[+-]?[0-9]+",
            "exp_dec": f"[+-]?[0-9]+[{self.dec_sep}][0-9]*[eE][+-]*[0-9]+",
            "exp_no_dec": "[+-]?[0-9]+[eE][+-]*[0-9]+",
        }

    def analyse_format(self):
        """
        Analyse the format.

        WHAT IT DOES:
            1) analyse the string to achieve a general classification
                (decimal, no decimal, exp notation)
            2) pass the string to an appropriate parsing function.

        RETURNS:
            the result of the parsing function:
                tuple with
                    format code to be used in f-string.
                    suited Python type for the number, int or float.
        """
        # 1. format definitions. key = general classification.

        # 2. analyse the format to find the general classification.
        gen_class, string = [], self.input_string.strip()
        for k, v in self.regexes.items():
            test = re.fullmatch(v, string)
            if test:
                gen_class.append(k)
        if not gen_class:
            raise TypeError("unknown format -->", string)
        elif len(gen_class) > 1:
            raise TypeError("ambiguous result -->", string, gen_class)

        # 3. based on the general classification, parse the string
        method_name = "_parse_" + str(gen_class[0])
        method = getattr(self, method_name, lambda *args: "Undefined Format!")
        return method(string)

    def _parse_dec(self, s):
        """Number is a decimal."""
        lst = s.split(self.dec_sep)
        result = "f" if not lst[1] else "." + str(len(lst[1])) + "f"
        result = "+" + result if "+" in lst[0] else result
        return (result, float)

    def _parse_no_dec(self, s):
        """Number is an integer."""
        result = "+d" if "+" in s else "d"
        return (result, int)

    def _parse_exp_dec(self, s):
        """Number is a decimal in exponential notation."""
        lst_dec = s.split(self.dec_sep)
        lst_e = lst_dec[1].upper().split("E")
        result = "." + str(len(lst_e[0])) + "E"
        result = result.lower() if "e" in s else result
        return (result, float)

    def _parse_exp_no_dec(self, s):
        """Number is in exponential notation but has no decimal points."""
        lst_e = s.upper().split("E")
        result = "+.0E" if "+" in lst_e[0] else ".0E"
        result = result.lower() if "e" in s else result
        return (result, float)


###############################################################################


def dec2str_stripped(
    num: float | list[float] | npt.NDArray[np.float64], dec_places: int = 3, strip: str = "right"
) -> str:
    """
    Convert floating point number to string, with zeros stripped.

    Parameters
    ----------
    num : float or list of float
        scalar or list of decimal numbers.
    dec_places : int, optional
        number of decimal places to return. defaults to 3.
    strip : string, optional
        what to strip. 'right' (default), 'left' or 'both'.

    Returns
    -------
    list of string.
        numbers formatted as strings according to specification (see kwargs).
    """
    _num: list[float]
    if isinstance(num, list):
        _num = num.copy()
    if isinstance(num, np.ndarray):
        _num = num.tolist()
    else:
        try:
            _num = list(num)  # type:ignore
        except TypeError:  # input was scalar
            _num = [num]  # type:ignore

    if not isinstance(dec_places, int) or int(dec_places) < 1:
        raise ValueError(f"kwarg dec_places must be integer > 1 (got {dec_places})")

    if strip == "right":
        return [f"{n:.{str(dec_places)}f}".rstrip("0") for n in _num]  # type:ignore
    if strip == "left":
        return [f"{n:.{str(dec_places)}f}".lstrip("0") for n in _num]  # type:ignore
    if strip == "both":
        return [f"{n:.{str(dec_places)}f}".strip("0") for n in _num]  # type:ignore
    raise ValueError(f"kwarg 'strip' must be 'right', 'left' or 'both' (got '{strip}')")



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Benchmark geodesic distance calculation methods.

This file was initially generated with the Windsurf Editor, to test the
GPT-4.1 model (free, with limited time).
"""

import time

import numpy as np
from pyproj import Geod
from rich import box
from rich.console import Console
from rich.table import Table

from pyfuppes.geo import geodesic_dist, haversine_dist


def geodesic_dist_pyproj(lat: np.ndarray, lon: np.ndarray, ellipsoid="WGS84") -> np.ndarray:
    """
    Calculate geodesic distance in km along lat/lon coordinates.
    Uses pyproj.Geod for vectorized computation.

    Returns
        array of step-wise distance for each pair of lat/lon coordinates
    """
    assert lat.shape[0] == lon.shape[0], "lat/lon must be of same length."

    geod = Geod(ellps=ellipsoid)
    # Compute distances between consecutive points
    lons1, lats1 = lon[:-1], lat[:-1]
    lons2, lats2 = lon[1:], lat[1:]
    # inv returns fwd_az, back_az, dist in meters
    _, _, dist = geod.inv(lons1, lats1, lons2, lats2)
    result = np.zeros(lat.shape, dtype=float)
    result[1:] = dist / 1000.0  # Convert meters to km

    return result


def benchmark_geo_distance(size=10000, seed=42, N=100):
    np.random.seed(seed)
    # Generate random coordinates within plausible ranges
    lat = np.random.uniform(-90, 90, size)
    lon = np.random.uniform(-180, 180, size)

    results = []

    # Haversine benchmark (average over N runs)
    hav_times = []
    hav_total = 0.0
    for _ in range(N):
        t0 = time.perf_counter()
        dist_hav = haversine_dist(lat, lon)
        t1 = time.perf_counter()
        hav_times.append(t1 - t0)
        hav_total += dist_hav
    results.append({
        "method": "haversine_dist",
        "avg_time": np.mean(hav_times),
        "total_dist": hav_total / N,
    })

    # Geodesic benchmark (average over N runs)
    geo_times = []
    geo_total = 0.0
    for _ in range(N):
        t0 = time.perf_counter()
        dist_geo = geodesic_dist(lat, lon)
        total_geo = np.sum(dist_geo)
        t1 = time.perf_counter()
        geo_times.append(t1 - t0)
        geo_total += total_geo
    results.append({
        "method": "geodesic_dist",
        "avg_time": np.mean(geo_times),
        "total_dist": geo_total / N,
    })

    # Geodesic (pyproj) benchmark (average over N runs)
    pyproj_times = []
    pyproj_total = 0.0
    for _ in range(N):
        t0 = time.perf_counter()
        dist_pyproj = geodesic_dist_pyproj(lat, lon)
        total_pyproj = np.sum(dist_pyproj)
        t1 = time.perf_counter()
        pyproj_times.append(t1 - t0)
        pyproj_total += total_pyproj
    results.append({
        "method": "geodesic_dist_pyproj",
        "avg_time": np.mean(pyproj_times),
        "total_dist": pyproj_total / N,
    })

    results.sort(key=lambda x: x["avg_time"])

    console = Console()
    table = Table(title=f"Geo Distance Benchmark ({size} points, {N} runs)", box=box.SIMPLE_HEAVY)
    table.add_column("Method", style="bold")
    table.add_column("Avg Time (s)", justify="right")
    table.add_column("Total Distance (km)", justify="right")
    fastest_time = results[0]["avg_time"]
    slowest_time = results[-1]["avg_time"]

    for res in results:
        style = (
            "green"
            if res["avg_time"] == fastest_time
            else ("red" if res["avg_time"] == slowest_time else "")
        )
        table.add_row(
            res["method"], f"{res['avg_time']:.6f}", f"{res['total_dist']:.3f}", style=style
        )

    console.print(table)


if __name__ == "__main__":
    benchmark_geo_distance(size=10000, N=10)



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Check monotonicity of Python lists and numpy 1D arrays."""

import numpy as np

###############################################################################


### Python list
def strictly_increasing(py_list: list) -> bool:
    """Check if elements of a list are strictly increasing."""
    return all(x < y for x, y in zip(py_list, py_list[1:], strict=False))


def strictly_decreasing(py_list: list) -> bool:
    """Check if elements of a list are strictly decreasing."""
    return all(x > y for x, y in zip(py_list, py_list[1:], strict=False))


def non_increasing(py_list: list) -> bool:
    """Check if elements of a list are increasing monotonically."""
    return all(x >= y for x, y in zip(py_list, py_list[1:], strict=False))


def non_decreasing(py_list: list) -> bool:
    """Check if elements of a list are decreasing monotonically."""
    return all(x <= y for x, y in zip(py_list, py_list[1:], strict=False))


### Numpy 1D array
def strict_inc_np(npndarr_1d: np.ndarray) -> np.bool_:
    """Check if elements of numpy 1D array are strictly increasing."""
    return np.all(np.diff(npndarr_1d) > 0)


def strict_dec_np(npndarr_1d: np.ndarray) -> np.bool_:
    """Check if elements of numpy 1D array are strictly decreasing."""
    return np.all(np.diff(npndarr_1d) < 0)


def non_inc_np(npndarr_1d: np.ndarray) -> np.bool_:
    """Check if elements of numpy 1D array are increasing monotonically."""
    return np.all(np.diff(npndarr_1d) <= 0)


def non_dec_np(npndarr_1d: np.ndarray) -> np.bool_:
    """Check if elements of numpy 1D array are decreasing monotonically."""
    return np.all(np.diff(npndarr_1d) >= 0)



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for plottools.py module."""

import unittest

import numpy as np
import numpy.ma as ma

# Add the parent directory to the path so we can import pyfuppes
from pyfuppes.plottools import get_plot_range, nticks_yrange


class TestPlotTools(unittest.TestCase):
    """Test class for plottools.py functions."""

    def test_get_plot_range_basic(self):
        """Test basic functionality of get_plot_range."""
        # Test with a simple list
        v = [1, 2, 3, 4, 5]
        result = get_plot_range(v)
        self.assertEqual(len(result), 2)
        self.assertLess(result[0], 1)  # Lower bound should be less than min value
        self.assertGreater(result[1], 5)  # Upper bound should be greater than max value

    def test_get_plot_range_numpy_array(self):
        """Test get_plot_range with numpy array."""
        v = np.array([1, 2, 3, 4, 5])
        result = get_plot_range(v)
        expected_offset = (abs(1) + abs(5)) / 2 * 5 / 100  # 5% of the range
        self.assertAlmostEqual(result[0], 1 - expected_offset)
        self.assertAlmostEqual(result[1], 5 + expected_offset)

    def test_get_plot_range_masked_array(self):
        """Test get_plot_range with masked array."""
        v = ma.array([1, 2, 3, 4, 5, 6], mask=[0, 0, 0, 0, 0, 1])
        result = get_plot_range(v)
        expected_offset = (abs(1) + abs(5)) / 2 * 5 / 100
        self.assertAlmostEqual(result[0], 1 - expected_offset)
        self.assertAlmostEqual(result[1], 5 + expected_offset)

    def test_get_plot_range_with_limits(self):
        """Test get_plot_range with min and max limits."""
        v = [1, 2, 3, 4, 5]
        result = get_plot_range(v, v_min_lim=1.5, v_max_lim=4.5)
        self.assertEqual(result[0], 1.5)  # Should be clamped to min limit
        self.assertEqual(result[1], 4.5)  # Should be clamped to max limit

    def test_get_plot_range_with_xrange(self):
        """Test get_plot_range with x range filtering."""
        x = [1, 2, 3, 4, 5]
        v = [10, 20, 30, 40, 50]
        result = get_plot_range(v, xrange=[2, 4], x=x)
        expected_offset = (abs(20) + abs(40)) / 2 * 5 / 100
        self.assertAlmostEqual(result[0], 20 - expected_offset)
        self.assertAlmostEqual(result[1], 40 + expected_offset)

    def test_get_plot_range_with_nan(self):
        """Test get_plot_range with NaN values."""
        v = np.array([1, 2, np.nan, 4, 5])
        result = get_plot_range(v)
        expected_offset = (v[0] + v[-1]) / 2 * 5 / 100
        self.assertAlmostEqual(result[0], 1 - expected_offset)
        self.assertAlmostEqual(result[1], 5 + expected_offset)

    def test_get_plot_range_empty_or_single(self):
        """Test get_plot_range with empty or single-value arrays."""
        v = np.array([])
        result = get_plot_range(v)
        self.assertEqual(result, (-1, 1))  # Default range for empty arrays

        v = np.array([42])
        result = get_plot_range(v)
        self.assertEqual(result, (-1, 1))  # Default range for single value arrays

    def test_nticks_yrange_basic(self):
        """Test basic functionality of nticks_yrange."""
        yrange = [1.3, 5.7]
        result = nticks_yrange(yrange, nticks=5)
        self.assertEqual(result, (0.0, 20.0))

    def test_nticks_yrange_custom_multiple(self):
        """Test nticks_yrange with custom multiple."""
        yrange = [1.3, 5.7]
        result = nticks_yrange(yrange, nticks=5, range_as_multiple_of=5)
        self.assertEqual(result, (0.0, 20.0))

        yrange = [1.3, 1.7]
        result = nticks_yrange(yrange, nticks=5, range_as_multiple_of=1)
        self.assertEqual(result, (1.0, 5.0))

    def test_nticks_yrange_negative_values(self):
        """Test nticks_yrange with negative values."""
        yrange = [-5.7, 1.3]
        result = nticks_yrange(yrange, nticks=5)
        self.assertEqual(result, (-10.0, 10.0))  # Should handle negative values correctly

    def test_nticks_yrange_exact_multiples(self):
        """Test nticks_yrange with values that are exact multiples."""
        yrange = [10.0, 30.0]
        result = nticks_yrange(yrange, nticks=3)
        self.assertEqual(result, (10.0, 30.0))  # Should keep the range as is


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
from datetime import UTC, datetime

import numpy as np
import pandas as pd
import xarray as xr

from pyfuppes import timeconversion


class TestTimeconv(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # to run before all tests
        pass

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    def test_to_list(self):
        self.assertEqual(timeconversion._to_list("hello"), (["hello"], True))
        self.assertEqual(
            timeconversion._to_list(datetime(2025, 5, 9)), ([datetime(2025, 5, 9)], True)
        )
        self.assertEqual(timeconversion._to_list([1, 2, 3]), ([1, 2, 3], False))
        self.assertEqual(timeconversion._to_list(np.array([4, 5, 6])), ([4, 5, 6], False))
        self.assertEqual(timeconversion._to_list(42), ([42], True))
        self.assertEqual(timeconversion._to_list(3.14), ([3.14], True))
        self.assertEqual(timeconversion._to_list([]), ([], False))
        self.assertEqual(timeconversion._to_list(np.array([])), ([], False))

    def test_xrtime_to_mdns(self):
        da = xr.DataArray(
            data=[1, 2, 3],
            dims=["Time"],
            coords={"Time": pd.date_range("2014-09-06", periods=3)},
        )
        t = timeconversion.xrtime_to_mdns(da)
        self.assertTrue(isinstance(t, np.ndarray))
        self.assertTrue(t.dtype.kind == "f")  # f --> floating point number
        self.assertTrue(issubclass(t.dtype.type, np.floating))
        self.assertListEqual([0.0, 86400.0, 172800.0], list(t))

    def test_dtstr_2_mdns(self):
        # no timezone
        t = ["2012-01-01T01:00:00", "2012-01-01T02:00:00"]
        f = "%Y-%m-%dT%H:%M:%S"
        result = list(map(int, timeconversion.dtstr_2_mdns(t, f)))
        self.assertEqual(result, [3600, 7200])
        # with timezone
        t = ["2012-01-01T01:00:00+02:00", "2012-01-01T02:00:00+02:00"]
        f = "%Y-%m-%dT%H:%M:%S%z"
        result = list(map(int, timeconversion.dtstr_2_mdns(t, f)))
        self.assertEqual(result, [3600, 7200])
        # varying UTC offset
        t = ["2012-01-01T01:00:00+02:00", "2012-01-01T02:00:00+01:00"]
        f = "%Y-%m-%dT%H:%M:%S%z"
        with self.assertRaises(AssertionError):
            _ = list(map(int, timeconversion.dtstr_2_mdns(t, f)))
        # zero case
        t = "2012-01-01T00:00:00+02:00"
        result = timeconversion.dtstr_2_mdns(t, f)
        self.assertEqual(int(result), 0)

    def test_dtobj_2_mdns(self):
        t = [datetime(2000, 1, 1, 1), datetime(2000, 1, 1, 2)]
        result = list(map(int, timeconversion.dtobj_2_mdns(t)))
        self.assertEqual(result, [3600, 7200])
        t = [
            datetime(2000, 1, 1, 1, tzinfo=UTC),
            datetime(2000, 1, 1, 2, tzinfo=UTC),
        ]
        result = list(map(int, timeconversion.dtobj_2_mdns(t)))
        self.assertEqual(result, [3600, 7200])

    def test_unixtime_2_mdns(self):
        t = [3600, 7200, 10800]
        result = list(map(int, timeconversion.unixtime_2_mdns(t)))
        self.assertEqual(result, t)

    def test_mdns_2_dtobj(self):
        t = [3600, 10800, 864000]
        ref = datetime(2020, 5, 15, tzinfo=UTC)
        result = list(map(int, timeconversion.mdns_2_dtobj(t, ref, posix=True)))
        self.assertEqual(result, [1589504400, 1589511600, 1590364800])

    def test_daysSince_2_dtobj(self):
        t0, off = datetime(2020, 5, 10), 10.5
        result = timeconversion.daysSince_2_dtobj(t0, off)
        self.assertEqual(result.hour, 12)
        self.assertEqual(result.day, 20)

    def test_dtstr_2_unix(self):
        result = timeconversion.dtstr_2_unixtime("2020-05-15", "%Y-%m-%d")
        self.assertAlmostEqual(result, datetime(2020, 5, 15, tzinfo=UTC).timestamp())


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest

from pyfuppes import numberstring


class TestNumberstring(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # to run before all tests
        print("\ntesting pyfuppes.numberstring...")

    @classmethod
    def tearDownClass(cls):
        # to run after all tests
        pass

    def setUp(self):
        # to run before each test
        pass

    def tearDown(self):
        # to run after each test
        pass

    def test_NumStr(self):
        valid = [
            "45",
            "45.",
            "3E5",
            "4E+5",
            "3E-3",
            "2.345E+7",
            "-7",
            "-45.3",
            "-3.4E3",
            " 12 ",
            "8.8e1",
            "+5.3",
            "+4.",
            "+10",
            "+2.3E121",
            "+4e-3",
            "-204E-9668",
            ".7",
            "+.7",
        ]
        for s in valid:
            result = numberstring.NumStr(s).analyse_format()
            # print(s, result[0])
            self.assertIsNotNone(result)
            self.assertIsInstance(result, tuple)

        invalid = ["tesT", "Test45", "7,7E2", "204-100", "."]
        for s in invalid:
            with self.assertRaises(TypeError):
                numberstring.NumStr(s).analyse_format()

    def test_NumStr_roundtrip(self):
        round_trip_ok = [
            "45",
            "45.3379\n",
            "-7",
            "-45.30",
            " 12 ",
            "+5.3",
            "+10",
        ]
        for s in round_trip_ok:
            result = numberstring.NumStr(s).analyse_format()
            fmt, value = result[0], result[1](s)
            self.assertEqual(s.strip(), f"{value:{fmt}}")

    def test_dec2str_stripped(self):
        numbers = [0.010701]
        self.assertEqual(
            "0.011",
            numberstring.dec2str_stripped(numbers, dec_places=3, strip="right")[0],
        )
        self.assertEqual(
            ".011",
            numberstring.dec2str_stripped(numbers, dec_places=3, strip="left")[0],
        )
        self.assertEqual(
            ".011",
            numberstring.dec2str_stripped(numbers, dec_places=3, strip="both")[0],
        )

        numbers = [1.0, 3.44532, 0.12011]
        self.assertEqual(
            ["1.", "3.445", "0.12"],
            numberstring.dec2str_stripped(numbers, dec_places=3, strip="right"),
        )
        self.assertEqual(
            ["1.000", "3.445", ".120"],
            numberstring.dec2str_stripped(numbers, dec_places=3, strip="left"),
        )
        self.assertEqual(
            ["1.", "3.445", ".12"],
            numberstring.dec2str_stripped(numbers, dec_places=3, strip="both"),
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Averaging, Binning, Mapping."""

from cmath import phase, rect
from copy import deepcopy
from math import atan2, cos, degrees, pi, radians, sin
from typing import NamedTuple

import numpy as np
import pandas as pd
from numba import njit  # type:ignore
from scipy.interpolate import interp1d  # type:ignore
from scipy.ndimage import uniform_filter1d  # type:ignore
from scipy.stats import circmean  # type:ignore

###############################################################################


def mean_angle(deg: list | np.ndarray | np.ma.masked_array) -> float:
    """
    Calculate a mean angle.

    input:
        deg - list or numpy array; values to average

    notes:
        - if input parameter deg contains NaN or is a numpy masked array, missing
          values will be removed before the calculation.
        - result is degrees between -180 and +180.

    Returns
    -------
        mean of deg (float)
    """
    if isinstance(deg, np.ma.masked_array):
        deg = deg.data
    elif isinstance(deg, np.ndarray):
        deg = deg[np.isfinite(deg)]

    if len(deg) == 0:
        return np.nan
    elif len(deg) == 1:
        return float(deg[0])

    return degrees(phase(sum(rect(1, radians(d)) for d in deg) / len(deg)))


###############################################################################


@njit
def mean_angle_numba(angles: np.ndarray) -> float:
    """
    mean_angle(), numba-JIT compiled.

    - input must be numpy array of type float!

    C version: https://rosettacode.org/wiki/Averages/Mean_angle
    """
    angles = angles[np.isfinite(angles)]
    if len(angles) == 0:
        return np.nan
    elif len(angles) == 1:
        return angles[0]

    size = len(angles)
    y_part = 0.0
    x_part = 0.0

    for i in range(size):
        x_part += cos(angles[i] * pi / 180)
        y_part += sin(angles[i] * pi / 180)

    return atan2(y_part / size, x_part / size) * 180 / pi


###############################################################################


def mean_angle_sc(deg: np.ndarray) -> float:
    """
    scipy.stats.circmean-based version of mean_angle().
    """
    if len(deg) == 0:
        return np.nan
    if len(deg) == 1:
        return deg[0]

    result = np.rad2deg(circmean(np.deg2rad(deg), nan_policy="omit"))
    if np.isclose(result, 0):
        return 0.0

    if result > 180:  # map to +-180 for consistency with other functions
        result -= 360

    return float(result)


###############################################################################


def mean_day_frac(dfr: list | np.ndarray | np.ma.masked_array, use_numba: bool = True) -> float:
    """
    Calculate a mean day fraction (0-1) with mean_angle function.

    the conversion to angle is necessary since day changes cannot be
      calculated as arithmetic mean.
    - dfr: day fraction, 0-1
    - if input parameter dfr contains NaN or is a numpy masked array, missing
      values will be removed before the calculation.
    """
    if isinstance(dfr, np.ma.masked_array):
        _dfr = dfr.data
    elif isinstance(dfr, np.ndarray):
        _dfr = dfr[np.isfinite(dfr)]
    else:
        _dfr = np.array(dfr, dtype=float)
        _dfr = _dfr[np.isfinite(_dfr)]

    # _dfr is now guaranteed to be of np.ndarray

    if len(_dfr) == 0:
        return np.nan
    elif len(_dfr) == 1:
        return _dfr[0]

    deg_mean = mean_angle_numba(_dfr * 360) if use_numba else mean_angle(_dfr * 360)

    if np.isclose(deg_mean, 0):
        return 0.0

    if deg_mean < 0:  # account for mean angle between -180 and +180
        deg_mean += 360

    return deg_mean / 360


###############################################################################


class TimeBins(NamedTuple):
    t_binned: np.ndarray
    bins: np.ndarray
    masked_bins: np.ndarray
    masked_vals: np.ndarray


def bin_t_10s(t: np.ndarray, force_t_range: bool = True, drop_empty: bool = True) -> TimeBins:
    """
    Bin a time axis to 10 s intervals around 5.

    Lower boundary included, upper boundary excluded (0. <= 5. < 10.).

    input:
        t - np.ndarray (time vector, unit=second, increasing monotonically)
    keywords:
        force_t_range (bool) - True enforces bins to fall within range of t
        drop_empty (bool) - False keeps empty bins alive

    Returns
    -------
        dict with binned time axis and bins, as returned by np.searchsorted()
    """
    if t.ndim != 1:
        raise TypeError("Please pass 1D array to function.")

    from pyfuppes.monotonicity import strict_inc_np

    if not strict_inc_np(t):
        raise ValueError("Input must be strictly increasing.")

    tmin, tmax = np.floor(t[0]), np.floor(t[-1])
    t_binned = np.arange((tmin - tmin % 10) + 5, (tmax - tmax % 10) + 6, 10)

    # if all values of t should fall WITHIN the range of t_binned:
    vmask = np.array([], dtype=bool)
    if force_t_range:
        if t_binned[0] < t[0]:
            t_binned = t_binned[1:]
        if t_binned[-1] > t[-1]:
            t_binned = t_binned[:-1]
        # check if values should be masked, e.g. if an element in t does not
        # fall into the bins
        vmask = (t < t_binned[0] - 5) | (t >= t_binned[-1] + 5)
        t = t[~vmask]

    bins = np.searchsorted(t_binned - 5, t, side="right")

    # if empty bins should be created, mask all bins that would have no
    # corresponding value in the dependent variable's data
    bmask = np.array([], dtype=bool)
    if drop_empty:
        t_binned = t_binned[np.bincount(bins - 1).astype(np.bool_)]
    else:
        bmask = np.ones(t_binned.shape).astype(np.bool_)
        bmask[bins - 1] = False

    return TimeBins(
        t_binned,
        bins,
        bmask,
        vmask,
    )


###############################################################################


@njit
def get_npnanmean(v: np.ndarray):
    """nan-mean, numba-JIT compiled."""
    return np.nanmean(v)


def bin_y_of_t(
    v: np.ndarray,
    bin_info: TimeBins,
    vmiss: float = np.nan,
    aggregation: str = "mean",
    use_numba: bool = True,
) -> np.ndarray:
    """
    Use the output of function "bin_time" or "bin_time_10s" to bin a variable 'v' that depends on a variable t.

    input:
        v - np.ndarray to be binned
        bin_info - config dict returned by bin_time() or bin_time_10s()
    keywords:
        vmiss (numeric) - missing value identifier, defaults to np.nan
        return_type (str) - how to bin, defaults to 'mean'. Options:
            'mean' - arithmetic mean
            'mean_day_frac' - mean of fractional day
            'mean_angle' - mean of angle
        use_numba (bool) - use njit'ed binning functions or not

    Returns
    -------
        v binned according to parameters in bin_info
    """
    if not any(
        v.dtype == np.dtype(t) for t in ("int16", "int32", "int64", "float16", "float32", "float64")
    ):
        raise TypeError("Please pass valid dtype, int or float.")

    # make a deep copy so that v is not modified on the way
    _v = deepcopy(v)

    # change dtype to float so we can use NaN
    if any(_v.dtype == np.dtype(t) for t in ("int16", "int32", "int64")):
        _v = _v.astype(float)

    _v[_v == vmiss] = np.nan

    # remove values that were masked (out of bin range)
    _v = _v[~bin_info.masked_vals]

    v_binned = []
    vd_bins = bin_info.bins

    if aggregation == "mean":
        if use_numba:
            v_binned = [get_npnanmean(_v[vd_bins == bin_no]) for bin_no in np.unique(vd_bins)]
        else:
            v_binned = [np.nanmean(_v[vd_bins == bin_no]) for bin_no in np.unique(vd_bins)]
            #
            # another option from <https://codereview.stackexchange.com/a/294677/206249>:
            #
            # _, inverse, counts = np.unique(vd_bins, return_inverse=True, return_counts=True)
            # v_binned = np.bincount(inverse, weights=_v) / counts
            #
    elif aggregation == "mean_day_frac":
        if use_numba:
            v_binned = [mean_day_frac(_v[vd_bins == bin_no]) for bin_no in np.unique(vd_bins)]
        else:
            v_binned = [
                mean_day_frac(_v[vd_bins == bin_no], use_numba=False)
                for bin_no in np.unique(vd_bins)
            ]
    elif aggregation == "mean_angle":
        if use_numba:
            v_binned = [mean_angle_numba(_v[vd_bins == bin_no]) for bin_no in np.unique(vd_bins)]
        else:
            v_binned = [mean_angle(_v[vd_bins == bin_no]) for bin_no in np.unique(vd_bins)]

    result = np.array(v_binned)

    # check if there are masked bins, i.e. empty bins. add them to the output if so.
    if bin_info.masked_bins.shape[0] > 0:
        tmp = np.ones(bin_info.masked_bins.shape)
        tmp[bin_info.masked_bins] = vmiss
        tmp[~bin_info.masked_bins] = result
        result = tmp

    # round to integers if input type was integer
    if any(v.dtype == np.dtype(t) for t in ("int16", "int32", "int64")):
        result = np.rint(result).astype(v.dtype)

    return result


###############################################################################

five_seconds = pd.Timedelta(seconds=5)


def bin_by_pdresample(
    t: np.ndarray,
    v: np.ndarray,
    t_unit: str = "s",
    rule: str = "10s",
    offset: pd.Timedelta | None = five_seconds,  # type: ignore
    force_t_range: bool = True,
    drop_empty: bool = True,
) -> pd.DataFrame:
    """
    Use pandas DataFrame method "resample" for binning along a time axis.

    Can only sample down, see also https://stackoverflow.com/q/66967998/10197418.

    Parameters
    ----------
    t : 1d array of float or int
        time axis / independent variable. unit: see keyword 't_unit'.
    v : 1d or 2d array corresponding to t
        dependent variable(s).
    rule : string, optional
        rule for resample method. The default is '10S'.
    offset : datetime.timedelta, optional
        offset to apply to the starting value.
        The default is timedelta(seconds=5).
    force_t_range : boolean, optional
        truncate new time axis to min/max of t. The default is True.
    drop_empty : boolean, optional
        drop empty bins that otherwise hold NaN. The default is True.

    Returns
    -------
    pandas DataFrame
        data binned (arithmetic mean) to resampled time axis.
    """
    d = {f"v_{i}": y for i, y in enumerate(v)} if isinstance(v, list) else {"v_0": v}

    df = pd.DataFrame(d, index=pd.to_datetime(t, unit=t_unit))
    df = df.resample(rule).mean()
    if offset:
        df.index = df.index + pd.tseries.frequencies.to_offset(offset)  # type:ignore

    df["t_binned"] = df.index.values.astype(np.int64) / 10**9

    if force_t_range:
        if df["t_binned"].iloc[0] < t[0]:
            df = df.drop(df.index[0])
        if df["t_binned"].iloc[-1] > t[-1]:
            df = df.drop(df.index[-1])

    df = df.drop(columns=["t_binned"]).set_index(df["t_binned"])

    if drop_empty:
        df = df.dropna(how="all")

    return df


###############################################################################


def bin_by_npreduceat(v: np.ndarray, nbins: int, ignore_nan: bool = True) -> np.ndarray:
    """
    Bin with numpy.add.reduceat (1D).

    ignores NaN or INF by default (finite elements only).
    if ignore_nan is set to False, the whole bin will be NaN if 1 or more NaNs
        fall within the bin.
    """
    bins = np.linspace(0, v.size, nbins + 1, True).astype(int)

    if ignore_nan:
        mask = np.isfinite(v)
        vn = np.where(~mask, 0, v)
        with np.errstate(invalid="ignore"):
            out = np.add.reduceat(vn, bins[:-1]) / np.add.reduceat(mask, bins[:-1])
    else:
        out = np.add.reduceat(v, bins[:-1]) / np.diff(bins)

    return out


###############################################################################


def np_mvg_avg(
    v: np.ndarray,
    N: int,
    ip_ovr_nan: bool = False,
    mode: str = "same",
    expand_edges: bool = True,
) -> np.ndarray:
    """
    Calculate moving average based on numpy convolution function.

    Parameters
    ----------
    v : 1d array
        data to average.
    N : integer
        number of samples per average.
    ip_ovr_nan : boolean, optional
        interpolate linearly using finite elements of v. The default is False.
    mode : string, optional
        config for np.convolve. The default is 'same'.
    expand_edges : bool, optional
        in case of mode='same', convolution gives incorrect results
        ("running-in effect") at edges. account for this by
        simply expanding the Nth value to the edges.

    Returns
    -------
    m_avg : 1d array
        averaged data.
    """
    if ip_ovr_nan:
        x = np.linspace(0, len(v) - 1, num=len(v))
        fip = interp1d(
            x[np.isfinite(v)],
            v[np.isfinite(v)],
            kind="linear",
            bounds_error=False,
            fill_value="extrapolate",  # type: ignore
        )
        v = fip(x)

    m_avg = np.convolve(v, np.ones((N,)) / N, mode=mode)  # type: ignore

    if expand_edges:
        m_avg[: N - 1], m_avg[-N:] = m_avg[N - 1], m_avg[-N]

    return m_avg


###############################################################################


def pd_mvg_avg(
    v: list | np.ndarray, N: int, ip_ovr_nan: bool = False, min_periods: int = 1
) -> np.ndarray:
    """
    Calculate moving average based on pandas dataframe rolling function.

    Parameters
    ----------
    v : 1d array
        data to average.
    N : integer
        number of samples per average.
    ip_ovr_nan : boolean, optional
        interpolate linearly using finite elements of v. The default is False.
    min_periods : TYPE, optional
        minimum number of values in averaging window. The default is 1.

    NOTE: automatically skips NaN (forms averages over windows with <N),
          unless minimum number of values in window is exceeded.

    Returns
    -------
    1d array
        averaged data.
    """
    N, min_periods = int(N), int(min_periods)

    min_periods = max(min_periods, 1)

    df = pd.DataFrame({"v": v})
    df["rollmean"] = df["v"].rolling(int(N), center=True, min_periods=min_periods).mean()
    if ip_ovr_nan:
        df["ip"] = df["rollmean"].interpolate()
        return df["ip"].values  # type: ignore

    return df["rollmean"].values  # type: ignore


###############################################################################


def sp_mvg_avg(v: np.ndarray, N: int, edges: str = "nearest") -> np.ndarray:
    """
    Use scipy's uniform_filter1d to calculate a moving average.

    See the docs at
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.uniform_filter1d.html
    Handles NaNs by removing them before interpolation.

    Parameters
    ----------
    v : np.ndarray
        data to average.
    N : int
        number of samples per average.
    edges : str, optional
        mode of uniform_filter1d (see docs at link above). The default is 'nearest'.
        {‘reflect’, ‘constant’, ‘nearest’, ‘mirror’, ‘wrap’}

    Returns
    -------
    avg : np.ndarray
        averaged data.
    """
    m = np.isfinite(v)
    avg = np.empty(v.shape)
    avg[~m] = np.nan
    avg[m] = uniform_filter1d(v[m], size=N, mode=edges)

    return avg


###############################################################################


def map_dependent(
    xref: np.ndarray, xcmp: np.ndarray, vcmp: np.ndarray, vmiss: float = np.nan
) -> np.ndarray:
    """
    Map a variable "vcmp" depending on variable "xcmp" to an independent variable "xref".

    Parameters
    ----------
    xref : np.ndarray, 1D
        reference / independent variable.
    xcmp : np.ndarray, 1D
        independent variable of vcmp.
    vcmp : np.ndarray, 1D
        dependent variable of xcmp.
    vmiss : int or float
        what should be inserted to specify missing values.

    Returns
    -------
    vmap : np.ndarray, 1D
        vcmp mapped to xref.
    """
    # which element of xref has a corresponding element in xcmp?
    m = np.isin(xref, xcmp)

    # prepare output
    vmap = np.empty(xref.shape, dtype=vcmp.dtype)
    # insert VMISS where xref has NO corresponding element
    vmap[~m] = vmiss

    # where corresponding elements exist, insert those from vcmp
    vmap[m] = np.take(vcmp, np.nonzero(np.isin(xcmp, xref)))[0]

    return vmap


###############################################################################


@njit
def calc_shift(
    arr: np.ndarray,
    step: float = 1,
    lower_bound: float = -2,
    upper_bound: float = 3,
    _tol: float = 1e-9,
) -> np.ndarray:
    """
    Calculate shift-values that, when added to arr, put the values of arr on a regular grid.
    Code gets numba-JIT compiled as the used ordinary for loop would make it very slow otherwise.

    Parameters
    ----------
    arr : np.ndarray
        irregularly gridded numbers.
    step : float, optional
        step size to expect in a regular grid. The default is 1.
    lower_bound : float, optional
        shift lower bound, included. The default is -2.
    upper_bound : float, optional
        shift upper bound, excluded. The default is 3.
    _tol : float, optional
        tolerance for divisibility check. The default is 1e-9.

    Returns
    -------
    shift : np.ndarray
        array with shift values.

    Examples
    --------
    >>> import numpy as np
    >>> arr = np.array([0.1, 1.2, 2.3, 3.1])
    >>> shift = calc_shift(arr, step=1.0)
    >>> np.allclose(shift, [-0.1, -0.2, -0.3, -0.1])
    True
    >>> # Verify the shifted array is on a regular grid
    >>> shifted = arr + shift
    >>> np.allclose(shifted, [0.0, 1.0, 2.0, 3.0])
    True
    """
    shift = np.zeros(arr.shape[0])
    # first element of arr must be evenly divisible by step, otherwise the first
    # element of shift must account for that
    m = arr[0] % step
    dy = abs(step - m)
    if not (m < _tol or dy < _tol):
        shift[0] = -(arr[0] % step)

    for i, v in enumerate(arr[1:]):
        offset = (arr[i] - v + shift[i]) + step
        if offset < lower_bound or offset >= upper_bound:
            offset = 0
        shift[i + 1] = offset

    return shift


###############################################################################

if __name__ == "__main__":
    import doctest

    doctest.testmod()



# SPDX-FileCopyrightText: 2025 Florian Obersteiner / KIT
# SPDX-FileContributor: Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""NA read/write."""

import os
from datetime import date
from pathlib import Path
from typing import IO, Any

###############################################################################

KEYS = [
    "NLHEAD",
    "ONAME",
    "ORG",
    "SNAME",
    "MNAME",
    "IVOL",
    "NVOL",
    "DATE",
    "RDATE",
    "DX",
    "XNAME",
    "NV",
    "VSCAL",
    "VMISS",
    "_VNAME",
    "NSCOML",
    "_SCOM",
    "NNCOML",
    "_NCOM",
    "_X",
    "_V",
    "_SRC",
    "_HEADER",
]


def na1001_cls_read(
    file: str | Path | IO,
    sep: str = " ",
    sep_data: str = "\t",
    auto_nncoml: bool = True,
    strip_lines: bool = True,
    rmv_repeated_seps: bool = False,
    vscale_vmiss_vertical: bool = False,
    vmiss_to_None: bool = False,
    ensure_ascii: bool = False,
    allow_emtpy_data: bool = False,
) -> dict:
    """
    Read NASA Ames 1001 formatted text file. Expected encoding is ASCII.

    See class method for detailled docstring.
    """
    na_1001: dict[str, Any] = {}
    if hasattr(file, "read"):  # if it has a read method, assume buffered IO
        na_1001["_SRC"] = "BaseIO"
        data = file.read()  # type: ignore
        if not isinstance(data, bytes):
            data = bytes(data, "utf-8")
    else:
        file = Path(file)  # type: ignore
        na_1001["_SRC"] = file.as_posix()
        with open(file, "rb") as f:
            data = f.read()

    # by definition, NASA Ames 1001 is pure ASCII. the following lines allow
    # to read files with other encodings; use with caution
    encodings = ("ascii",) if ensure_ascii else ("ascii", "utf-8", "cp1252", "latin-1")

    decoded: str | None = None
    for enc in encodings:
        try:
            decoded = data.decode(enc)
        except ValueError:  # invalid encoding, try next
            pass
        else:
            if enc != "ascii":
                print(f"warning: non-ascii encoding '{enc}' used in file {na_1001['_SRC']}")
            break  # found a working encoding
    if not decoded:
        raise ValueError(f"could not decode input (ASCII-only: {ensure_ascii})")

    file_content = decoded.split("\n")

    if strip_lines:
        for i, line in enumerate(file_content):
            file_content[i] = line.strip()

    if rmv_repeated_seps:
        for i, line in enumerate(file_content):
            while sep + sep in line:
                line = line.replace(sep + sep, sep)  # noqa: PLW2901
            file_content[i] = line

    tmp: list[int] = list(map(int, file_content[0].split()))
    assert len(tmp) == 2, f"invalid format in line 1: '{file_content[0]}'"
    assert tmp[0] >= 15, f"NASA Ames FFI 1001 has a least 15 header lines (specified: {tmp[0]})"
    assert tmp[1] == 1001, f"reader is for FFI 1001 only, got {tmp[1]}"

    nlhead: int = tmp[0]
    na_1001["NLHEAD"] = nlhead

    header = file_content[:nlhead]
    data = file_content[nlhead:]
    if all(x == "" for x in data) or data == ["\n"]:
        data = None

    if not allow_emtpy_data:
        assert data, "no data found."

    na_1001["ONAME"] = header[1]
    na_1001["ORG"] = header[2]
    na_1001["SNAME"] = header[3]
    na_1001["MNAME"] = header[4]

    tmp = list(map(int, header[5].split()))
    assert len(tmp) == 2, f"invalid format in line 6: '{header[5]}'"
    na_1001["IVOL"], na_1001["NVOL"] = tmp[0], tmp[1]

    tmp = list(map(int, header[6].split()))
    assert len(tmp) == 6, f"invalid format line 7: '{header[6]}'"

    # check for valid date in line 7 (yyyy mm dd)
    assert date(*tmp[:3]) <= date(*tmp[3:6]), (
        f"RDATE must be greater or equal to DATE, have DATE {date(*tmp[:3])}, RDATE {date(*tmp[3:6])}"
    )
    na_1001["DATE"], na_1001["RDATE"] = tmp[:3], tmp[3:6]

    # DX check if the line contains a decimal separator; if so use float else int
    na_1001["DX"] = float(header[7]) if "." in header[7] else int(header[7])
    na_1001["XNAME"] = header[8]

    n_vars = int(header[9])
    na_1001["NV"] = n_vars

    if vscale_vmiss_vertical:
        offset = n_vars * 2
        na_1001["VSCAL"] = header[10 : 10 + n_vars]
        na_1001["VMISS"] = header[10 + n_vars : 10 + n_vars * 2]
    else:
        offset = 2
        na_1001["VSCAL"] = header[10].split()
        na_1001["VMISS"] = header[11].split()

    assert len(na_1001["VSCAL"]) == na_1001["NV"], (
        f"number of elements in VSCAL (have: {len(na_1001['VSCAL'])}) must match number of variables specified ({na_1001['NV']})"
    )
    assert len(na_1001["VMISS"]) == na_1001["NV"], (
        f"number of elements in VMISS (have: {len(na_1001['VMISS'])}) must match number of variables specified ({na_1001['NV']})"
    )
    assert n_vars == len(na_1001["VSCAL"]) == len(na_1001["VMISS"]), (
        "VSCAL, VMISS and NV must have equal number of elements"
    )

    na_1001["_VNAME"] = header[10 + offset : 10 + n_vars + offset]

    nscoml = int(header[10 + n_vars + offset])
    na_1001["NSCOML"] = nscoml
    if nscoml > 0:  # read special comment if nscoml>0
        na_1001["_SCOM"] = header[n_vars + 11 + offset : n_vars + nscoml + 11 + offset]
    else:
        na_1001["_SCOM"] = ""

    msg = "nscoml not equal n elements in list na_1001['_SCOM']"
    assert nscoml == len(na_1001["_SCOM"]), msg

    # read normal comment if nncoml>0
    if auto_nncoml is True:
        nncoml = nlhead - (n_vars + nscoml + 12 + offset)
    else:
        nncoml = int(header[n_vars + nscoml + 11 + offset])
    na_1001["NNCOML"] = nncoml

    if nncoml > 0:
        na_1001["_NCOM"] = header[
            n_vars + nscoml + 12 + offset : n_vars + nscoml + nncoml + 12 + offset
        ]
    else:
        na_1001["_NCOM"] = ""

    msg = "nncoml not equal n elements in list na_1001['_NCOM']"
    assert nncoml == len(na_1001["_NCOM"]), msg

    msg = "nlhead must be equal to nncoml + nscoml + n_vars + 14"
    assert nncoml + nscoml + n_vars + 14 == nlhead, msg

    # done with header, we can set HEADER variable now
    na_1001["_HEADER"] = header

    # continue with variables
    na_1001["_X"] = []  # holds independent variable
    na_1001["_V"] = [[] for _ in range(n_vars)]  # list for each dependent variable

    if data is not None:
        for ix, line in enumerate(data):
            if line == "" or line == "\n":  # skip empty lines or trailing newline
                continue

            parts = line.rsplit(sep=sep_data)
            assert len(parts) == n_vars + 1, (
                f"invalid number of parameters in line {ix + nlhead + 1}, have {len(parts)} ({parts}), want {n_vars + 1}"
            )

            na_1001["_X"].append(parts[0].strip())
            if vmiss_to_None:
                for j in range(n_vars):
                    na_1001["_V"][j].append(
                        parts[j + 1].strip()
                        if parts[j + 1].strip() != na_1001["VMISS"][j]
                        else None
                    )
            else:
                for j in range(n_vars):
                    na_1001["_V"][j].append(parts[j + 1].strip())

    return na_1001


###############################################################################


def na1001_cls_write(
    file_path: str | Path,
    na_1001: dict,
    sep: str = " ",
    sep_data: str = "\t",
    overwrite: int | bool = 0,
    encoding: str = "utf-8",
    verbose: bool = False,
) -> int:
    """
    Write content of na1001 class instance to file in NASA Ames 1001 format. Encoding is ASCII.

    See class method for detailled docstring.
    """
    verboseprint = print if verbose else lambda *a, **k: None

    # check if directory exists, create if not.
    if not os.path.isdir(os.path.dirname(file_path)):
        os.mkdir(os.path.dirname(file_path))

    # check if file exists, act according to overwrite keyword
    if os.path.isfile(file_path):
        if not overwrite:
            verboseprint(
                f"write failed: {file_path} already exists.\nset overwrite keyword to overwrite."
            )
            return 0  # write failed / forbidden
        write = 2  # overwriting
    write = 1  # normal writing

    # check n variables and comment lines; adjust values if incorrect
    n_vars_named = len(na_1001["_VNAME"])
    n_vars_data = len(na_1001["_V"])
    if n_vars_named != n_vars_data:
        raise ValueError(
            f"NA error: n vars in V and VNAME not equal, {n_vars_data} vs. {n_vars_named}!"
        )

    if n_vars_named - na_1001["NV"] != 0:
        verboseprint("NA output: NV corrected")
        na_1001["NV"] = n_vars_named

    nscoml_is = len(na_1001["_SCOM"])
    if (nscoml_is - na_1001["NSCOML"]) != 0:
        verboseprint("NA output: NSCOML corrected")
        na_1001["NSCOML"] = nscoml_is

    nncoml_is = len(na_1001["_NCOM"])
    if (nncoml_is - na_1001["NNCOML"]) != 0:
        verboseprint("NA output: NNCOML corrected")
        na_1001["NNCOML"] = nncoml_is

    nlhead_is = 14 + n_vars_named + nscoml_is + nncoml_is
    if (nlhead_is - na_1001["NLHEAD"]) != 0:
        verboseprint("NA output: NLHEAD corrected")
        na_1001["NLHEAD"] = nlhead_is

    # begin the actual writing process
    with open(file_path, "w", encoding=encoding) as file_obj:
        block = str(na_1001["NLHEAD"]) + sep + "1001\n"
        file_obj.write(block)

        block = str(na_1001["ONAME"]) + "\n"
        file_obj.write(block)

        block = str(na_1001["ORG"]) + "\n"
        file_obj.write(block)

        block = str(na_1001["SNAME"]) + "\n"
        file_obj.write(block)

        block = str(na_1001["MNAME"]) + "\n"
        file_obj.write(block)

        block = str(na_1001["IVOL"]) + sep + str(na_1001["NVOL"]) + "\n"
        file_obj.write(block)

        # dates: assume "yyyy m d" in tuple
        block = (
            f"{na_1001['DATE'][0]:04d}{sep}{na_1001['DATE'][1]:02d}{sep}{na_1001['DATE'][2]:02d}{sep}"
            f"{na_1001['RDATE'][0]:04d}{sep}{na_1001['RDATE'][1]:02d}{sep}{na_1001['RDATE'][2]:02d}{sep}\n"
        )
        file_obj.write(block)

        file_obj.write(f"{na_1001['DX']}" + "\n")

        # obsolete: CARIBIC
        # file_obj.write(sep_com.join(na_1001["XNAME"]) + "\n")
        file_obj.write(na_1001["XNAME"] + "\n")

        n_vars = na_1001["NV"]  # get number of variables
        block = str(n_vars) + "\n"
        file_obj.write(block)

        line = ""
        for i in range(n_vars):
            line += str(na_1001["VSCAL"][i]) + sep
        line = line[0:-1] if line.endswith("\n") else line[0:-1] + "\n"
        file_obj.write(line)

        line = ""
        for i in range(n_vars):
            line += str(na_1001["VMISS"][i]) + sep
        line = line[0:-1] if line.endswith("\n") else line[0:-1] + "\n"
        file_obj.write(line)

        block = na_1001["_VNAME"]
        for i in range(n_vars):
            file_obj.write(block[i] + "\n")

        nscoml = na_1001["NSCOML"]  # get number of special comment lines
        line = str(nscoml) + "\n"
        file_obj.write(line)

        block = na_1001["_SCOM"]
        for i in range(nscoml):
            file_obj.write(block[i] + "\n")

        nncoml = na_1001["NNCOML"]  # get number of normal comment lines
        line = str(nncoml) + "\n"
        file_obj.write(line)

        block = na_1001["_NCOM"]
        for i in range(nncoml):
            file_obj.write(block[i] + "\n")

        for i, x in enumerate(na_1001["_X"]):
            line = str(x) + sep_data
            for j in range(n_vars):
                line += str(na_1001["_V"][j][i]) + sep_data
            file_obj.write(line[0:-1] + "\n")

    return write
